import inspect

from pydantic import BaseConfig
from pydantic.fields import ModelField


def model_field_from_param(param: inspect.Parameter) -> ModelField:
    return ModelField.infer(
        name=param.name,
        value=param.default if param.default is not param.empty else ...,
        annotation=param.annotation,
        class_validators={},
        config=BaseConfig,
    )



import inspect
import sys
from dataclasses import dataclass
from typing import Any, ClassVar, Dict, Iterable, List, Optional, Tuple, cast

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

import starlette.types
from pydantic.fields import ModelField
from starlette.requests import HTTPConnection

from xpresso.binders._extractors.api import ParameterExtractorMarker
from xpresso.binders._extractors.params.base import (
    ParameterExtractorBase,
    get_basic_param_info,
)
from xpresso.binders._extractors.utils import grouped, is_mapping_like, is_sequence_like
from xpresso.typing import Some


def collect_sequence(value: str) -> List[str]:
    if not value:
        return []
    return [v for v in value.split(",") if v]


def collect_object(value: str) -> Dict[str, str]:
    if not value:
        return {}
    return dict(cast(Iterable[Tuple[str, str]], grouped([v for v in value.split(",")])))


def collect_scalar(value: str) -> str:
    return value


class Extractor(Protocol):
    def __call__(self, value: str) -> Any:
        ...


def get_extractor(explode: bool, field: ModelField) -> Extractor:
    if is_sequence_like(field):
        if explode is True:
            raise ValueError(  # pragma: no cover
                "To deserialize array cookies, you must use Cookie(explode=False)"
            )
        return collect_sequence
    if is_mapping_like(field):
        if explode is True:
            raise ValueError(  # pragma: no cover
                "To deserialize object cookies, you must use Cookie(explode=False)"
            )
        return collect_object
    # single item
    return collect_scalar


@dataclass(frozen=True)
class CookieParameterExtractor(ParameterExtractorBase):
    extractor: Extractor
    in_: ClassVar[str] = "cookie"

    async def extract(
        self,
        scope: starlette.types.Scope,
        receive: starlette.types.Receive,
        send: starlette.types.Send,
        connection: HTTPConnection,
    ) -> Any:
        param = connection.cookies.get(self.name, None)
        if param is not None:
            return await self.validate(
                Some(self.extractor(param)), scope, receive, send
            )
        return await self.validate(None, scope, receive, send)


@dataclass(frozen=True)
class CookieParameterExtractorMarker(ParameterExtractorMarker):
    alias: Optional[str]
    explode: bool
    in_: ClassVar[str] = "cookie"

    def register_parameter(self, param: inspect.Parameter) -> CookieParameterExtractor:
        field, name, loc = get_basic_param_info(param, self.alias, self.in_)
        extractor = get_extractor(field=field, explode=self.explode)
        return CookieParameterExtractor(
            field=field, loc=loc, name=name, extractor=extractor
        )



import inspect
import typing
from dataclasses import dataclass

from di.typing import get_markers_from_parameter

from xpresso._utils.typing import model_field_from_param
from xpresso.binders._openapi_providers.api import (
    ModelNameMap,
    OpenAPIBody,
    OpenAPIBodyMarker,
    Schemas,
)
from xpresso.binders.dependants import BodyBinderMarker
from xpresso.openapi import models


@dataclass(frozen=True)
class FieldOpenAPIBase(OpenAPIBody):
    name: str
    include_in_schema: bool
    field_openapi: OpenAPIBody

    def get_encoding(self) -> typing.Optional[models.Encoding]:
        return self.field_openapi.get_encoding()

    def get_field_name(self) -> str:
        return self.name


class OpenAPIField(FieldOpenAPIBase):
    def get_schema(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.Schema:
        return self.field_openapi.get_schema(
            model_name_map=model_name_map, schemas=schemas
        )


class OpenAPIRepeatedField(FieldOpenAPIBase):
    def get_schema(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.Schema:
        return models.Schema(
            type="array",
            items=self.field_openapi.get_schema(
                model_name_map=model_name_map, schemas=schemas
            ),
        )


@dataclass(frozen=True)
class OpenAPIFieldMarkerBase(OpenAPIBodyMarker):
    alias: typing.Optional[str]
    include_in_schema: bool
    cls: typing.ClassVar[
        typing.Union[typing.Type[OpenAPIRepeatedField], typing.Type[OpenAPIField]]
    ]

    def register_parameter(self, param: inspect.Parameter) -> OpenAPIBody:
        field_marker: typing.Optional[OpenAPIBodyMarker] = None
        for marker in get_markers_from_parameter(param):
            if isinstance(marker, BodyBinderMarker):
                if marker.openapi_marker is not self:
                    # the outermost marker must be the field marker (us)
                    # so the first one that isn't us is the inner marker
                    field_marker = marker.openapi_marker
        if field_marker is None:
            raise TypeError(
                "No field marker found"
                "\n You must include a valid field marker using ExtractField[AsJson[...]]"
                " or Annotated[..., Json(), Field()]"
            )
        field_openapi = field_marker.register_parameter(param)
        name = self.alias or model_field_from_param(param).alias
        return self.cls(
            name=name,
            field_openapi=field_openapi,
            include_in_schema=self.include_in_schema,
        )


class OpenAPIFieldMarker(OpenAPIFieldMarkerBase):
    cls = OpenAPIField


class OpenAPIRepeatedFieldMarker(OpenAPIFieldMarkerBase):
    cls = OpenAPIRepeatedField



import sys

if sys.version_info < (3, 8):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict


from di import BaseContainer
from starlette.responses import Response


class _XpressoASGIExtension(TypedDict):
    container: BaseContainer
    response_sent: bool


class XpressoASGIExtension(_XpressoASGIExtension, total=False):
    response: Response



from typing import Any, Callable, Iterable, Type

from starlette.datastructures import UploadFile as StarletteUploadFile


class UploadFile(StarletteUploadFile):
    @classmethod
    def __get_validators__(cls: Type["UploadFile"]) -> Iterable[Callable[..., Any]]:
        return iter([])



from starlette import status as status
from starlette.responses import Response

from xpresso.applications import App
from xpresso.binders._param_functions import (
    ByContentType,
    CookieParam,
    ExtractField,
    ExtractRepeatedField,
    File,
    Form,
    FormEncodedField,
    FormField,
    FromCookie,
    FromFile,
    FromFormData,
    FromFormField,
    FromHeader,
    FromJson,
    FromMultipart,
    FromPath,
    FromQuery,
    HeaderParam,
    Json,
    Multipart,
    PathParam,
    QueryParam,
    RepeatedFormField,
)
from xpresso.datastructures import UploadFile
from xpresso.dependencies.models import Dependant
from xpresso.exceptions import HTTPException
from xpresso.requests import Request
from xpresso.routing.operation import Operation
from xpresso.routing.pathitem import Path
from xpresso.routing.router import Router
from xpresso.routing.websockets import WebSocketRoute
from xpresso.security._functions import Security
from xpresso.websockets import WebSocket

__all__ = (
    "Operation",
    "Path",
    "QueryParam",
    "HeaderParam",
    "CookieParam",
    "PathParam",
    "Json",
    "ExtractField",
    "ExtractRepeatedField",
    "Form",
    "File",
    "Multipart",
    "Dependant",
    "App",
    "Router",
    "UploadFile",
    "Security",
    "FromCookie",
    "FromMultipart",
    "FromFormField",
    "FromFormData",
    "FromHeader",
    "FromJson",
    "FormEncodedField",
    "FromPath",
    "FromQuery",
    "HTTPException",
    "RepeatedFormField",
    "FormField",
    "FromFile",
    "status",
    "Request",
    "Response",
    "ByContentType",
    "WebSocketRoute",
    "WebSocket",
)



import typing

import starlette.requests
import starlette.responses
import starlette.routing
import starlette.types
import starlette.websockets
from di import AsyncExecutor, BaseContainer, ConcurrentAsyncExecutor
from di.api.dependencies import DependantBase
from di.api.executor import AsyncExecutorProtocol
from di.api.providers import DependencyProvider as Endpoint
from di.api.solved import SolvedDependant
from di.dependant import JoinedDependant

import xpresso._utils.asgi_scope_extension as asgi_scope_extension
from xpresso.dependencies.models import Dependant


class _WebSocketRoute:
    __slots__ = ("dependant", "executor")

    def __init__(
        self,
        dependant: SolvedDependant[typing.Any],
        executor: AsyncExecutorProtocol,
    ) -> None:
        self.dependant = dependant
        self.executor = executor

    async def __call__(
        self,
        scope: starlette.types.Scope,
        receive: starlette.types.Receive,
        send: starlette.types.Send,
    ) -> None:
        ws = starlette.websockets.WebSocket(scope=scope, receive=receive, send=send)
        conn = starlette.requests.HTTPConnection(scope=scope, receive=receive)
        values: typing.Dict[typing.Any, typing.Any] = {
            starlette.websockets.WebSocket: ws,
            starlette.requests.HTTPConnection: conn,
            starlette.types.Scope: scope,
            starlette.types.Receive: receive,
            starlette.types.Send: send,
        }
        xpresso_scope: asgi_scope_extension.XpressoASGIExtension = scope["extensions"][
            "xpresso"
        ]
        async with xpresso_scope["container"].enter_scope("operation") as container:
            await container.execute_async(
                self.dependant,
                values=values,
                executor=self.executor,
            )


class WebSocketRoute(starlette.routing.WebSocketRoute):
    def __init__(
        self,
        path: str,
        endpoint: Endpoint,
        *,
        name: typing.Optional[str] = None,
        dependencies: typing.Optional[typing.Sequence[Dependant]] = None,
        execute_dependencies_concurrently: bool = False,
    ) -> None:
        super().__init__(  # type: ignore
            path=path,
            endpoint=endpoint,
            name=name,  # type: ignore[arg-type]
        )
        self.endpoint = endpoint
        self.dependencies = dependencies or []
        self.execute_dependencies_concurrently = execute_dependencies_concurrently

    def solve(
        self,
        container: BaseContainer,
        dependencies: typing.List[DependantBase[typing.Any]],
    ) -> None:
        self.dependant = container.solve(
            JoinedDependant(
                Dependant(self.endpoint, scope="operation"),
                siblings=[*dependencies, *(self.dependencies or ())],
            )
        )
        executor: AsyncExecutorProtocol
        if self.execute_dependencies_concurrently:
            executor = ConcurrentAsyncExecutor()
        else:
            executor = AsyncExecutor()
        self.app = _WebSocketRoute(
            dependant=self.dependant,
            executor=executor,
        )



import typing
from contextlib import asynccontextmanager

import starlette.types
from di import AsyncExecutor, BaseContainer
from di.api.dependencies import DependantBase
from di.api.providers import DependencyProviderType
from starlette.applications import Starlette
from starlette.datastructures import State
from starlette.middleware import Middleware
from starlette.middleware.errors import ServerErrorMiddleware
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse, Response
from starlette.routing import BaseRoute
from starlette.routing import Route as StarletteRoute

from xpresso._utils.asgi_scope_extension import XpressoASGIExtension
from xpresso._utils.routing import visit_routes
from xpresso.dependencies.models import Dependant
from xpresso.dependencies.utils import register_framework_dependencies
from xpresso.exception_handlers import (
    http_exception_handler,
    validation_exception_handler,
)
from xpresso.exceptions import HTTPException, RequestValidationError
from xpresso.middleware.exceptions import ExceptionMiddleware
from xpresso.openapi import models as openapi_models
from xpresso.openapi._builder import SecurityModels, genrate_openapi
from xpresso.openapi._html import get_swagger_ui_html
from xpresso.routing.pathitem import Path
from xpresso.routing.router import Router
from xpresso.routing.websockets import WebSocketRoute
from xpresso.security._dependants import Security

ExceptionHandler = typing.Callable[[Request, typing.Type[BaseException]], Response]


class App(Starlette):
    router: Router
    middleware_stack: starlette.types.ASGIApp
    openapi: typing.Optional[openapi_models.OpenAPI] = None
    _debug: bool
    state: State
    exception_handlers: typing.Mapping[
        typing.Union[int, typing.Type[Exception]], ExceptionHandler
    ]
    user_middleware: typing.Sequence[Middleware]

    def __init__(
        self,
        routes: typing.Optional[typing.Sequence[BaseRoute]] = None,
        *,
        container: typing.Optional[BaseContainer] = None,
        dependencies: typing.Optional[typing.List[Dependant]] = None,
        debug: bool = False,
        middleware: typing.Optional[typing.Sequence[Middleware]] = None,
        exception_handlers: typing.Optional[
            typing.Dict[
                typing.Union[int, typing.Type[Exception]],
                ExceptionHandler,
            ]
        ] = None,
        lifespan: typing.Optional[DependencyProviderType[None]] = None,
        openapi_version: str = "3.0.3",
        title: str = "API",
        description: typing.Optional[str] = None,
        version: str = "0.1.0",
        openapi_url: typing.Optional[str] = "/openapi.json",
        docs_url: typing.Optional[str] = "/docs",
        servers: typing.Optional[typing.Iterable[openapi_models.Server]] = None,
    ) -> None:
        routes = list(routes or [])
        routes.extend(
            self._get_doc_routes(
                openapi_url=openapi_url,
                docs_url=docs_url,
            )
        )
        self._debug = debug
        self.state = State()
        self.exception_handlers = (
            {} if exception_handlers is None else dict(exception_handlers)
        )

        self.container = container or BaseContainer(
            scopes=("app", "connection", "operation")
        )
        register_framework_dependencies(self.container)
        self._setup_run = False

        @asynccontextmanager
        async def lifespan_ctx(app: Starlette) -> typing.AsyncGenerator[None, None]:
            self._setup()
            self._setup_run = True
            original_container = self.container
            async with self.container.enter_scope("app") as container:
                self.container = container
                if lifespan is not None:
                    await container.execute_async(
                        self.container.solve(Dependant(call=lifespan, scope="app")),
                        executor=AsyncExecutor(),
                    )
                try:
                    yield
                finally:
                    # make this cm reentrant for testing purposes
                    self.container = original_container
                    self._setup_run = False

        self.router = Router(routes, lifespan=lifespan_ctx, dependencies=dependencies)
        self.user_middleware = [] if middleware is None else list(middleware)
        self.middleware_stack = self.build_middleware_stack()  # type: ignore
        self.add_exception_handler(RequestValidationError, validation_exception_handler)  # type: ignore
        self.add_exception_handler(HTTPException, http_exception_handler)  # type: ignore
        self.openapi_version = openapi_version
        self.openapi_info = openapi_models.Info(
            title=title,
            version=version,
            description=description,
        )
        self.servers = servers

    def build_middleware_stack(self) -> starlette.types.ASGIApp:
        debug = self.debug
        error_handler = None
        exception_handlers = {}

        for key, value in self.exception_handlers.items():
            if key in (500, Exception):
                error_handler = value
            else:
                exception_handlers[key] = value

        middleware = (
            Middleware(ServerErrorMiddleware, handler=error_handler, debug=debug),
            *self.user_middleware,
            Middleware(ExceptionMiddleware, handlers=exception_handlers, debug=debug),
        )

        app = self.router
        for cls, options in reversed(middleware):
            app = cls(app=app, **options)
        return app

    async def __call__(
        self,
        scope: starlette.types.Scope,
        receive: starlette.types.Receive,
        send: starlette.types.Send,
    ) -> None:
        self._setup()
        if scope["type"] == "http" or scope["type"] == "websocket":
            extensions = scope.get("extensions", None) or {}
            scope["extensions"] = extensions
            xpresso_scope = extensions.get("xpresso", None)
            if xpresso_scope is None:
                async with self.container.enter_scope("connection") as container:
                    xpresso_asgi_extension: XpressoASGIExtension = {
                        "container": container,
                        "response_sent": False,
                    }
                    extensions["xpresso"] = xpresso_asgi_extension
                    await super().__call__(scope, receive, send)
                    xpresso_asgi_extension["response_sent"] = True
                    return
        await super().__call__(scope, receive, send)

    def _setup(self) -> None:
        if self._setup_run:
            return
        for route in visit_routes([self.router]):
            dependencies: typing.List[DependantBase[typing.Any]] = []
            for router in route.routers:
                if isinstance(router, Router):
                    dependencies.extend(router.dependencies)
            if isinstance(route.route, Path):
                for operation in route.route.operations.values():
                    operation.solve(
                        dependencies=[
                            *dependencies,
                            *route.route.dependencies,
                            *operation.dependencies,
                        ],
                        container=self.container,
                    )
            if isinstance(route.route, WebSocketRoute):
                route.route.solve(
                    dependencies=[
                        *dependencies,
                        *route.route.dependencies,
                    ],
                    container=self.container,
                )

    async def get_openapi(self) -> openapi_models.OpenAPI:
        return genrate_openapi(
            version=self.openapi_version,
            info=self.openapi_info,
            servers=self.servers,
            router=self.router,
            security_models=await self.gather_security_models(),
        )

    async def gather_security_models(self) -> SecurityModels:
        security_dependants: typing.List[Security] = []
        for route in visit_routes([self.router]):
            if isinstance(route.route, Path):
                for operation in route.route.operations.values():
                    dependant = operation.dependant
                    assert dependant is not None
                    for subdependant in dependant.dag:
                        if isinstance(subdependant, Security):
                            security_dependants.append(subdependant)
        executor = AsyncExecutor()
        return {
            sec_dep: await self.container.execute_async(
                self.container.solve(sec_dep),
                executor=executor,
            )
            for sec_dep in security_dependants
        }

    def _get_doc_routes(
        self,
        openapi_url: typing.Optional[str],
        docs_url: typing.Optional[str],
    ) -> typing.Iterable[BaseRoute]:
        routes: typing.List[BaseRoute] = []

        if openapi_url:
            openapi_url = openapi_url

            async def openapi(req: Request) -> JSONResponse:
                if self.openapi is None:
                    self.openapi = await self.get_openapi()
                res = JSONResponse(self.openapi.dict(exclude_none=True, by_alias=True))
                return res

            routes.append(
                StarletteRoute(
                    path=openapi_url, endpoint=openapi, include_in_schema=False
                )
            )
        if openapi_url and docs_url:

            openapi_url = openapi_url

            async def swagger_ui_html(req: Request) -> HTMLResponse:
                root_path: str = req.scope.get("root_path", "").rstrip("/")
                full_openapi_url = root_path + openapi_url  # type: ignore[operator]
                return get_swagger_ui_html(
                    openapi_url=full_openapi_url,
                    title=self.openapi_info.title + " - Swagger UI",
                    oauth2_redirect_url=None,
                    init_oauth=None,
                )

            routes.append(
                StarletteRoute(
                    path=docs_url,
                    endpoint=swagger_ui_html,
                    include_in_schema=False,
                )
            )

        return routes



from __future__ import annotations

import inspect
import typing

from di.api.dependencies import DependantBase

from xpresso.binders._extractors.api import (
    BodyExtractor,
    BodyExtractorMarker,
    ParameterExtractor,
    ParameterExtractorMarker,
)
from xpresso.binders._openapi_providers.api import (
    OpenAPIBody,
    OpenAPIBodyMarker,
    OpenAPIParameter,
    OpenAPIParameterMarker,
)
from xpresso.dependencies.models import Dependant


class ParameterBinderMarker(Dependant):
    def __init__(
        self,
        *,
        in_: str,
        extractor_marker: ParameterExtractorMarker,
        openapi_marker: OpenAPIParameterMarker,
    ) -> None:
        super().__init__(call=None, scope="connection")
        self.in_ = in_
        self.extractor_marker = extractor_marker
        self.openapi_marker = openapi_marker

    def register_parameter(self, param: inspect.Parameter) -> DependantBase[typing.Any]:
        return ParameterBinder(
            in_=self.in_,
            openapi=self.openapi_marker.register_parameter(param),
            extractor=self.extractor_marker.register_parameter(param),
        )


class ParameterBinder(Dependant):
    __slots__ = (
        "in_",
        "alias",
        "openapi",
    )

    def __init__(
        self,
        in_: str,
        openapi: OpenAPIParameter,
        extractor: ParameterExtractor,
    ):
        self.in_ = in_
        self.openapi = openapi
        super().__init__(call=extractor.extract, scope="connection")

    @property
    def cache_key(self) -> typing.Tuple[str, str]:
        # As per the spec, parameters are identified
        # by name and location
        return (self.openapi.name, self.in_)


class BodyBinderMarker(Dependant):
    def __init__(
        self,
        *,
        extractor_marker: BodyExtractorMarker,
        openapi_marker: OpenAPIBodyMarker,
    ) -> None:
        super().__init__(call=None, scope="connection")
        self.extractor_marker = extractor_marker
        self.openapi_marker = openapi_marker

    def register_parameter(self, param: inspect.Parameter) -> BodyBinder:
        return BodyBinder(
            openapi=self.openapi_marker.register_parameter(param),
            extractor=self.extractor_marker.register_parameter(param),
        )


class BodyBinder(Dependant):
    __slots__ = ("openapi", "extractor")

    def __init__(
        self,
        *,
        openapi: OpenAPIBody,
        extractor: BodyExtractor,
    ) -> None:
        super().__init__(call=extractor.extract_from_request, scope="connection")
        self.openapi = openapi
        self.extractor = extractor



import inspect
import sys
import typing
from dataclasses import dataclass

if sys.version_info < (3, 9):
    from typing_extensions import Annotated, get_args, get_origin
else:
    from typing import Annotated, get_origin, get_args

from di.typing import get_markers_from_parameter

from xpresso._utils.typing import model_field_from_param
from xpresso.binders._openapi_providers.api import (
    ModelNameMap,
    OpenAPIBody,
    OpenAPIBodyMarker,
    Schemas,
)
from xpresso.binders.dependants import BodyBinderMarker
from xpresso.openapi import models as openapi_models


@dataclass(frozen=True)
class OpenAPIContentTypeDiscriminated(OpenAPIBody):
    sub_body_providers: typing.Mapping[str, OpenAPIBody]
    description: typing.Optional[str]
    required: typing.Optional[bool]
    include_in_schema = True

    def get_openapi(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> openapi_models.RequestBody:
        return openapi_models.RequestBody(
            description=self.description,
            required=self.required,
            content={
                media_type: provider.get_media_type_object(model_name_map, schemas)
                for media_type, provider in self.sub_body_providers.items()
            },
        )


@dataclass(frozen=True)
class OpenAPIContentTypeDiscriminatedMarker(OpenAPIBodyMarker):
    description: typing.Optional[str]

    def register_parameter(self, param: inspect.Parameter) -> OpenAPIBody:
        field = model_field_from_param(param)
        if field.required is False:
            required = False
        else:
            required = True

        sub_body_providers: typing.Dict[str, OpenAPIBody] = {}

        annotation = param.annotation
        origin = get_origin(annotation)
        assert origin is Annotated
        annotation = next(iter(get_args(annotation)))
        origin = get_origin(annotation)
        if origin is not typing.Union:
            raise TypeError("Unioned bodies must be a Union of simple bodies")
        args = get_args(annotation)
        for arg in args:
            sub_body_param = inspect.Parameter(
                name=param.name,
                kind=param.kind,
                annotation=arg,
                default=param.default,
            )
            marker: typing.Optional[BodyBinderMarker] = None
            for param_marker in get_markers_from_parameter(sub_body_param):
                if isinstance(param_marker, BodyBinderMarker):
                    marker = param_marker
                    break
            if marker is None:
                raise TypeError(f"Type annotation is missing body marker: {arg}")
            sub_body_openapi = marker.openapi_marker
            provider = sub_body_openapi.register_parameter(sub_body_param)
            if provider.include_in_schema:
                media_type = provider.get_media_type()
                sub_body_providers[media_type] = provider
        return OpenAPIContentTypeDiscriminated(
            sub_body_providers=sub_body_providers,
            description=self.description,
            required=None if required else False,
        )



from __future__ import annotations

import sys
import typing

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

import di
from di.api.providers import DependencyProvider

T = typing.TypeVar("T")


Scope = Literal["app", "connection", "operation"]


class Dependant(di.Dependant[typing.Any]):
    __slots__ = ()

    def __init__(
        self,
        call: typing.Optional[DependencyProvider] = None,
        scope: Scope = "connection",
        use_cache: bool = True,
        wire: bool = True,
        sync_to_thread: bool = False,
    ) -> None:
        super().__init__(
            call=call,
            scope=scope,
            use_cache=use_cache,
            wire=wire,
            sync_to_thread=sync_to_thread,
        )



from xpresso.security._api_key import APIKeyCookie, APIKeyHeader, APIKeyQuery
from xpresso.security._functions import Security
from xpresso.security._http import (
    HTTPAuthorizationCredentials,
    HTTPBase,
    HTTPBasic,
    HTTPBasicCredentials,
    HTTPBearer,
    HTTPDigest,
)
from xpresso.security._oauth2 import (
    OAuth2,
    OAuth2AuthorizationCodeBearer,
    OAuth2PasswordBearer,
    OAuth2PasswordRequestForm,
    OAuth2PasswordRequestFormStrict,
    SecurityScopes,
)
from xpresso.security._open_id_connect_url import OpenIdConnect

__all__ = (
    "APIKeyCookie",
    "APIKeyCookie",
    "APIKeyHeader",
    "APIKeyQuery",
    "HTTPAuthorizationCredentials",
    "HTTPBasic",
    "HTTPBasicCredentials",
    "HTTPBearer",
    "HTTPDigest",
    "OAuth2",
    "OAuth2AuthorizationCodeBearer",
    "OAuth2PasswordBearer",
    "OAuth2PasswordRequestForm",
    "OAuth2PasswordRequestFormStrict",
    "Security",
    "SecurityScopes",
    "OpenIdConnect",
    "HTTPBase",
)



import typing

import starlette.applications
import starlette.background
import starlette.datastructures
import starlette.requests
import starlette.responses
import starlette.routing
import starlette.status
import starlette.types
from di import AsyncExecutor, BaseContainer, ConcurrentAsyncExecutor
from di.api.dependencies import DependantBase
from di.api.executor import AsyncExecutorProtocol
from di.api.providers import DependencyProvider as Endpoint
from di.api.solved import SolvedDependant
from di.dependant import JoinedDependant

import xpresso._utils.asgi_scope_extension as asgi_scope_extension
import xpresso.binders.dependants as param_dependants
import xpresso.openapi.models as openapi_models
from xpresso.dependencies.models import Dependant
from xpresso.encoders.api import Encoder
from xpresso.encoders.json import JsonableEncoder
from xpresso.responses import Responses


class _OperationApp:
    __slots__ = ("dependant", "executor", "response_factory", "response_encoder")

    def __init__(
        self,
        dependant: SolvedDependant[typing.Any],
        executor: AsyncExecutorProtocol,
        response_factory: typing.Callable[[typing.Any], starlette.responses.Response],
        response_encoder: Encoder,
    ) -> None:
        self.dependant = dependant
        self.executor = executor
        self.response_factory = response_factory
        self.response_encoder = response_encoder

    async def __call__(
        self,
        scope: starlette.types.Scope,
        receive: starlette.types.Receive,
        send: starlette.types.Send,
    ) -> None:
        values: typing.Dict[typing.Any, typing.Any] = {
            starlette.requests.Request: starlette.requests.Request(
                scope=scope, receive=receive, send=send
            ),
            starlette.requests.HTTPConnection: starlette.requests.HTTPConnection(
                scope=scope, receive=receive
            ),
            starlette.background.BackgroundTasks: starlette.background.BackgroundTasks(),
            starlette.types.Scope: scope,
            starlette.types.Receive: receive,
            starlette.types.Send: send,
        }
        xpresso_scope: asgi_scope_extension.XpressoASGIExtension = scope["extensions"][
            "xpresso"
        ]
        async with xpresso_scope["container"].enter_scope("operation") as container:
            endpoint_return = await container.execute_async(
                self.dependant,
                values=values,
                executor=self.executor,
            )
            if isinstance(endpoint_return, starlette.responses.Response):
                response = endpoint_return
            else:
                response = self.response_factory(self.response_encoder(endpoint_return))
            xpresso_scope["response"] = response
        await response(scope, receive, send)


class Operation(starlette.routing.BaseRoute):
    def __init__(
        self,
        endpoint: Endpoint,
        *,
        # OpenAPI parameters
        include_in_schema: bool = True,
        tags: typing.Optional[typing.Sequence[str]] = None,
        summary: typing.Optional[str] = None,
        description: typing.Optional[str] = None,
        deprecated: typing.Optional[bool] = None,
        operation_id: typing.Optional[bool] = None,
        servers: typing.Optional[typing.Sequence[openapi_models.Server]] = None,
        external_docs: typing.Optional[openapi_models.ExternalDocumentation] = None,
        responses: typing.Optional[Responses] = None,
        # xpresso params
        name: typing.Optional[str] = None,
        dependencies: typing.Optional[typing.Sequence[Dependant]] = None,
        execute_dependencies_concurrently: bool = False,
        response_factory: typing.Callable[
            [typing.Any], starlette.responses.Response
        ] = starlette.responses.JSONResponse,
        response_encoder: Encoder = JsonableEncoder(),
        sync_to_thread: bool = True,
    ) -> None:
        self._app: typing.Optional[starlette.types.ASGIApp] = None
        self.endpoint = endpoint
        self.tags = list(tags or [])
        self.summary = summary
        self.description = description
        self.deprecated = deprecated
        self.operation_id = operation_id
        self.servers = list(servers or [])
        self.external_docs = external_docs
        self.responses = dict(responses or {})
        self.dependencies = list(dependencies or [])
        self.execute_dependencies_concurrently = execute_dependencies_concurrently
        self.response_factory = response_factory
        self.response_encoder = response_encoder
        self.include_in_schema = include_in_schema
        self.name: str = starlette.routing.get_name(endpoint) if name is None else name  # type: ignore
        self.sync_to_thread = sync_to_thread

    async def handle(
        self,
        scope: starlette.types.Scope,
        receive: starlette.types.Receive,
        send: starlette.types.Send,
    ) -> None:
        if self._app is None:
            raise RuntimeError("Operation cannot be used outside of a Xpresso App")
        return await self._app(scope, receive, send)

    def solve(
        self,
        container: BaseContainer,
        dependencies: typing.List[DependantBase[typing.Any]],
    ) -> None:
        self.dependant = container.solve(
            JoinedDependant(
                Dependant(
                    self.endpoint, scope="operation", sync_to_thread=self.sync_to_thread
                ),
                siblings=[*dependencies, *(self.dependencies or ())],
            )
        )
        flat = self.dependant.get_flat_subdependants()
        bodies = [dep for dep in flat if isinstance(dep, param_dependants.BodyBinder)]
        if len(bodies) > 1:
            raise ValueError("There can only be 1 top level body per operation")
        executor: AsyncExecutorProtocol
        if self.execute_dependencies_concurrently:
            executor = ConcurrentAsyncExecutor()
        else:
            executor = AsyncExecutor()
        self._app = _OperationApp(
            dependant=self.dependant,
            executor=executor,
            response_encoder=self.response_encoder,
            response_factory=self.response_factory,
        )

    def url_path_for(
        self, name: str, **path_params: str
    ) -> starlette.datastructures.URLPath:
        if path_params:
            raise starlette.routing.NoMatchFound()
        if name != self.name:
            raise starlette.routing.NoMatchFound()
        return starlette.datastructures.URLPath("/")

    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, type(self)) and self.endpoint is __o.endpoint



import typing
from dataclasses import dataclass

from starlette.routing import BaseRoute, Mount
from starlette.routing import Router as StarletteRouter

from xpresso.responses import Responses
from xpresso.routing.pathitem import Path
from xpresso.routing.router import Router as XpressoRouter


@dataclass(frozen=True)
class VisitedRoute:
    path: str
    routers: typing.List[StarletteRouter]
    route: BaseRoute
    tags: typing.List[str]
    responses: Responses


def visit_routes(
    routers: typing.List[StarletteRouter],
    path: typing.Optional[str] = None,
    tags: typing.Optional[typing.List[str]] = None,
    responses: typing.Optional[Responses] = None,
) -> typing.Generator[VisitedRoute, None, None]:
    path = path or ""
    tags = tags or []
    responses = responses or {}
    router = next(iter(reversed(routers)), None)
    assert router is not None
    for route in typing.cast(typing.Iterable[BaseRoute], router.routes):
        if isinstance(route, Mount) and isinstance(route.app, StarletteRouter):
            child_tags = tags
            child_responses = responses
            if isinstance(route.app, XpressoRouter):
                child_tags = child_tags + route.app.tags
                child_responses = {**child_responses, **route.app.responses}
            yield from visit_routes(
                routers=routers + [route.app],
                path=path + route.path,
                tags=child_tags,
                responses=child_responses,
            )
        elif hasattr(route, "path"):
            route_path: str = route.path  # type: ignore
            child_tags = tags
            child_responses = responses
            if isinstance(route, Path):
                child_tags = child_tags + route.tags
                child_responses = {**child_responses, **route.responses}
            yield VisitedRoute(
                path=path + route_path,
                routers=routers,
                route=route,
                tags=child_tags,
                responses=child_responses,
            )



import sys
from typing import Generic, TypeVar

if sys.version_info < (3, 9):
    from typing_extensions import Annotated as Annotated  # noqa: F401
else:
    from typing import Annotated as Annotated  # noqa: F401

T = TypeVar("T")


class Some(Generic[T]):
    __slots__ = ("value",)

    def __init__(self, value: T) -> None:
        self.value = value

    def __repr__(self) -> str:
        return f"Some({repr(self.value)})"



import dataclasses
from collections import defaultdict
from enum import Enum
from pathlib import PurePath
from types import GeneratorType
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)

from pydantic import BaseModel
from pydantic.json import ENCODERS_BY_TYPE

from xpresso.encoders.api import Encoder
from xpresso.typing import Some

_SetIntStr = Set[Union[int, str]]
_DictIntStrAny = Dict[Union[int, str], Any]


def _generate_encoders_by_class_tuples(
    type_encoder_map: Dict[Any, Callable[[Any], Any]]
) -> Dict[Callable[[Any], Any], Tuple[Any, ...]]:
    encoders_by_class_tuples: Dict[Callable[[Any], Any], Tuple[Any, ...]] = defaultdict(
        tuple
    )
    for type_, encoder in type_encoder_map.items():
        encoders_by_class_tuples[encoder] += (type_,)
    return encoders_by_class_tuples


encoders_by_class_tuples = _generate_encoders_by_class_tuples(ENCODERS_BY_TYPE)


class JsonableEncoder(Encoder):
    def __init__(
        self,
        include: Optional[Union[_SetIntStr, _DictIntStrAny]] = None,
        exclude: Optional[Union[_SetIntStr, _DictIntStrAny]] = None,
        by_alias: bool = True,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        custom_encoder: Optional[Dict[Any, Callable[[Any], Any]]] = None,
    ) -> None:
        if include is not None:
            include = set(include)
        if exclude is not None:
            exclude = set(exclude)
        self.include = include
        self.exclude = exclude
        self.by_alias = by_alias
        self.exclude_unset = exclude_unset
        self.exclude_defaults = exclude_defaults
        self.exclude_none = exclude_none
        self.custom_encoder = custom_encoder or {}

    def apply_custom_encoder(
        self, obj: Any, custom_encoder: Optional[Dict[Any, Callable[[Any], Any]]]
    ) -> Optional[Some[Any]]:
        if custom_encoder:
            key = custom_encoder.get(type(obj), None)
            if key is not None:
                return Some(key(obj))
        return None

    def __call__(
        self, obj: Any, custom_encoder: Dict[Any, Callable[[Any], Any]] = {}
    ) -> Any:
        custom_encoder = {**self.custom_encoder, **custom_encoder}
        if isinstance(obj, BaseModel):
            encoder = getattr(obj.__config__, "json_encoders", {})
            if custom_encoder:
                encoder.update(custom_encoder)
            obj_dict = obj.dict(
                include=self.include,  # type: ignore # in Pydantic
                exclude=self.exclude,  # type: ignore # in Pydantic
                by_alias=self.by_alias,
                exclude_unset=self.exclude_unset,
                exclude_none=self.exclude_none,
                exclude_defaults=self.exclude_defaults,
            )
            if "__root__" in obj_dict:
                obj_dict = obj_dict["__root__"]
            return self(obj_dict, custom_encoder=encoder)
        if dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)
        if isinstance(obj, Enum):
            return obj.value
        if isinstance(obj, PurePath):
            return str(obj)
        if isinstance(obj, (str, int, float, type(None))):
            return obj
        if isinstance(obj, dict):
            encoded_dict: Dict[Any, Any] = {}
            for key, value in cast(Mapping[Any, Any], obj).items():
                if (value is not None or not self.exclude_none) and (
                    (self.include and key in self.include)
                    or not self.exclude
                    or key not in self.exclude
                ):
                    encoded_key = self(key, custom_encoder=custom_encoder)
                    encoded_value = self(value, custom_encoder=custom_encoder)
                    encoded_dict[encoded_key] = encoded_value
            return encoded_dict
        if isinstance(obj, (list, set, frozenset, GeneratorType, tuple)):
            encoded_list: List[Any] = []
            for item in cast(Sequence[Any], obj):
                encoded_list.append(self(item, custom_encoder=custom_encoder))
            return encoded_list

        custom = self.apply_custom_encoder(obj, custom_encoder=custom_encoder)
        if isinstance(custom, Some):
            return custom.value

        if type(obj) in ENCODERS_BY_TYPE:
            return ENCODERS_BY_TYPE[type(obj)](obj)
        for encoder, classes_tuple in encoders_by_class_tuples.items():
            if isinstance(obj, classes_tuple):
                return encoder(obj)

        errors: List[Exception] = []
        try:
            data = dict(obj)
        except Exception as e:
            errors.append(e)
            try:
                data = vars(obj)
            except Exception as e:
                errors.append(e)
                raise ValueError(errors)
        return self(data, custom_encoder=custom_encoder)



import inspect
import typing
from dataclasses import dataclass

from pydantic.fields import ModelField
from pydantic.schema import field_schema, get_flat_models_from_field, get_model_name_map

from xpresso._utils.typing import model_field_from_param
from xpresso.binders._openapi_providers.api import (
    ModelNameMap,
    OpenAPIBody,
    OpenAPIBodyMarker,
    Schemas,
)
from xpresso.openapi import models as openapi_models
from xpresso.openapi.constants import REF_PREFIX


@dataclass(frozen=True)
class OpenAPIFormField(OpenAPIBody):
    alias: typing.Optional[str]
    style: str
    explode: bool
    name: str
    field: ModelField
    include_in_schema: bool

    def get_field_name(self) -> str:
        return self.name

    def get_schema(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> openapi_models.Schema:
        model_name_map.update(
            get_model_name_map(
                get_flat_models_from_field(
                    self.field,
                    model_name_map.keys(),  # type: ignore[arg-type]
                )
            )
        )
        schema, refs, _ = field_schema(
            self.field,
            by_alias=True,
            ref_prefix=REF_PREFIX,
            model_name_map=model_name_map,
        )
        schemas.update(refs)
        return openapi_models.Schema(**schema)

    def get_encoding(self) -> typing.Optional[openapi_models.Encoding]:
        return openapi_models.Encoding(
            contentType=None,
            style=self.style,
            explode=self.explode,
        )


@dataclass(frozen=True)
class OpenAPIFormFieldMarker(OpenAPIBodyMarker):
    alias: typing.Optional[str]
    style: str
    explode: bool
    include_in_schema: bool

    def register_parameter(self, param: inspect.Parameter) -> OpenAPIBody:
        field = model_field_from_param(param)
        name = self.alias or field.alias
        return OpenAPIFormField(
            alias=self.alias,
            style=self.style,
            explode=self.explode,
            name=name,
            field=field,
            include_in_schema=self.include_in_schema,
        )



from xpresso.binders._extractors.api import (
    BodyExtractor,
    BodyExtractorMarker,
    ParameterExtractor,
    ParameterExtractorMarker,
)
from xpresso.binders._openapi_providers.api import (
    ModelNameMap,
    OpenAPIBody,
    OpenAPIBodyMarker,
    OpenAPIParameter,
    OpenAPIParameterMarker,
    Schemas,
)

__all__ = (
    "BodyExtractor",
    "BodyExtractorMarker",
    "ParameterExtractor",
    "ParameterExtractorMarker",
    "OpenAPIBody",
    "OpenAPIBodyMarker",
    "OpenAPIParameter",
    "OpenAPIParameterMarker",
    "ModelNameMap",
    "Schemas",
)



import typing

import starlette.applications
import starlette.background
import starlette.datastructures
import starlette.exceptions
import starlette.requests
import starlette.responses
import starlette.routing
import starlette.status
import starlette.types

from xpresso.dependencies.models import Dependant
from xpresso.responses import Responses


def _not_supported(method: str) -> typing.Callable[..., typing.Any]:
    def raise_error(*args: typing.Any, **kwargs: typing.Any) -> typing.NoReturn:
        raise NotImplementedError(
            f"Use of Router.{method} is deprecated."
            " Use Router(routes=[...]) instead."
        )

    return raise_error


class Router(starlette.routing.Router):
    routes: typing.List[starlette.routing.BaseRoute]

    def __init__(
        self,
        routes: typing.Sequence[starlette.routing.BaseRoute],
        *,
        redirect_slashes: bool = True,
        default: typing.Optional[starlette.types.ASGIApp] = None,
        lifespan: typing.Optional[
            typing.Callable[
                [starlette.applications.Starlette], typing.AsyncContextManager[None]
            ]
        ] = None,
        dependencies: typing.Optional[typing.List[Dependant]] = None,
        tags: typing.Optional[typing.List[str]] = None,
        responses: typing.Optional[Responses] = None,
    ) -> None:
        super().__init__(  # type: ignore
            routes=list(routes),
            redirect_slashes=redirect_slashes,
            default=default,  # type: ignore
            lifespan=lifespan,  # type: ignore
        )
        self.dependencies = list(dependencies or [])
        self.tags = list(tags or [])
        self.responses = dict(responses or {})

    mount = _not_supported("mount")
    host = _not_supported("host")
    add_route = _not_supported("add_route")
    add_websocket_route = _not_supported("add_websocket_route")
    route = _not_supported("route")
    websocket_route = _not_supported("websocket_route")
    add_event_handler = _not_supported("add_event_handler")
    on_event = _not_supported("on_event")



import inspect
from dataclasses import dataclass
from typing import Any, ClassVar, Optional

import starlette.types
from pydantic.error_wrappers import ErrorWrapper
from starlette.requests import HTTPConnection

from xpresso.binders._extractors.api import ParameterExtractor, ParameterExtractorMarker
from xpresso.binders._extractors.exceptions import InvalidSerialization
from xpresso.binders._extractors.form_utils import Extractor, get_extractor
from xpresso.binders._extractors.params.base import (
    ParameterExtractorBase,
    get_basic_param_info,
)
from xpresso.exceptions import RequestValidationError, WebSocketValidationError

ERRORS = {
    "webscoket": WebSocketValidationError,
    "http": RequestValidationError,
}


@dataclass(frozen=True)
class QueryParameterExtractor(ParameterExtractorBase):
    extractor: Extractor
    in_: ClassVar[str] = "query"

    async def extract(
        self,
        scope: starlette.types.Scope,
        receive: starlette.types.Receive,
        send: starlette.types.Send,
        connection: HTTPConnection,
    ) -> Any:
        try:
            extracted = self.extractor(
                name=self.name, params=connection.query_params.multi_items()
            )
        except InvalidSerialization:
            raise ERRORS[scope["type"]](
                [
                    ErrorWrapper(
                        exc=TypeError("Data is not a valid URL encoded query"),
                        loc=tuple(("query", self.name)),
                    )
                ]
            )
        return await self.validate(extracted, scope, receive, send)


@dataclass(frozen=True)
class QueryParameterExtractorMarker(ParameterExtractorMarker):
    alias: Optional[str]
    explode: bool
    style: str
    in_: ClassVar[str] = "query"

    def register_parameter(self, param: inspect.Parameter) -> ParameterExtractor:
        if self.style == "deepObject" and not self.explode:
            # no such thing in the spec
            raise ValueError("deepObject can only be used with explode=True")
        field, name, loc = get_basic_param_info(param, self.alias, self.in_)
        extractor = get_extractor(style=self.style, explode=self.explode, field=field)
        name = self.alias or field.alias
        return QueryParameterExtractor(
            field=field, loc=loc, name=name, extractor=extractor
        )



import inspect
import typing
from dataclasses import dataclass, field

from pydantic.fields import (
    SHAPE_FROZENSET,
    SHAPE_ITERABLE,
    SHAPE_LIST,
    SHAPE_SEQUENCE,
    SHAPE_SET,
    SHAPE_TUPLE_ELLIPSIS,
    ModelField,
)

from xpresso._utils.typing import model_field_from_param
from xpresso.binders._openapi_providers.api import (
    OpenAPIParameter,
    OpenAPIParameterMarker,
)
from xpresso.binders._openapi_providers.utils import parse_examples
from xpresso.openapi import models as openapi_models

Examples = typing.Optional[typing.Mapping[str, openapi_models.Example]]

SEQUENCE_LIKE_SHAPES = (
    SHAPE_LIST,
    SHAPE_SET,
    SHAPE_SEQUENCE,
    SHAPE_ITERABLE,
    SHAPE_TUPLE_ELLIPSIS,
    SHAPE_FROZENSET,
)


@dataclass(frozen=True)
class OpenAPIParameterBase(OpenAPIParameter):
    name: str
    in_: str
    required: bool
    style: str
    explode: bool
    description: typing.Optional[str]
    deprecated: typing.Optional[bool]
    include_in_schema: bool
    examples: Examples = field(compare=False)
    field: ModelField = field(compare=False)


@dataclass(frozen=True)
class OpenAPIParameterMarkerBase(OpenAPIParameterMarker):
    alias: typing.Optional[str]
    style: str
    explode: bool
    description: typing.Optional[str]
    deprecated: typing.Optional[bool]
    include_in_schema: bool
    examples: Examples
    in_: typing.ClassVar[str]
    cls: typing.ClassVar[typing.Type[OpenAPIParameterBase]]
    must_be_required: typing.ClassVar[bool] = False

    def register_parameter(self, param: inspect.Parameter) -> OpenAPIParameterBase:
        field = model_field_from_param(param)
        name = self.alias or field.alias
        if field.required is False and self.must_be_required:
            raise TypeError(
                f"{self.in_.title()} parameters MUST be required and MUST NOT have default values"
            )
        return self.cls(
            name=name,
            in_=self.in_,
            required=(field.required is not False)
            and (field.shape not in SEQUENCE_LIKE_SHAPES),
            field=field,
            style=self.style,
            explode=self.explode,
            description=self.description,
            deprecated=self.deprecated,
            include_in_schema=self.include_in_schema,
            examples=parse_examples(self.examples) if self.examples else None,
        )



from __future__ import annotations

import typing

from di import BaseContainer
from starlette.background import BackgroundTasks
from starlette.requests import HTTPConnection, Request
from starlette.websockets import WebSocket

from xpresso.dependencies.models import Dependant

T = typing.TypeVar("T")


def register_framework_dependencies(container: BaseContainer):
    container.register_by_type(
        Dependant(Request, scope="connection", wire=False),
        Request,
    )
    container.register_by_type(
        Dependant(Request, scope="connection", wire=False),
        Request,
    )
    container.register_by_type(
        Dependant(
            HTTPConnection,
            scope="connection",
            wire=False,
        ),
        HTTPConnection,
    )
    container.register_by_type(
        Dependant(
            WebSocket,
            scope="connection",
            wire=False,
        ),
        WebSocket,
    )
    container.register_by_type(
        Dependant(
            BackgroundTasks,
            scope="connection",
            wire=False,
        ),
        BackgroundTasks,
    )



import inspect
import typing
from dataclasses import dataclass

from pydantic.error_wrappers import ErrorWrapper
from starlette.datastructures import FormData

from xpresso._utils.typing import model_field_from_param
from xpresso.binders._extractors.api import BodyExtractor, BodyExtractorMarker
from xpresso.binders._extractors.exceptions import InvalidSerialization
from xpresso.binders._extractors.form_utils import (
    Extractor,
    UnexpectedFileReceived,
    get_extractor,
)
from xpresso.exceptions import RequestValidationError
from xpresso.typing import Some


@dataclass(frozen=True)
class FormFieldBodyExtractor(BodyExtractor, BodyExtractorMarker):
    alias: typing.Optional[str]
    style: str
    explode: bool
    name: str
    extractor: Extractor

    async def extract_from_form(
        self, form: FormData, *, loc: typing.Iterable[typing.Union[int, str]]
    ) -> typing.Optional[Some[typing.Any]]:
        try:
            return self.extractor(name=self.name, params=form.multi_items())
        except InvalidSerialization:
            raise RequestValidationError(
                [
                    ErrorWrapper(
                        exc=TypeError("Data is not a valid URL encoded form"),
                        loc=tuple((*loc, self.name)),
                    )
                ]
            )
        except UnexpectedFileReceived as exc:
            raise RequestValidationError(
                [
                    ErrorWrapper(
                        exc=exc,
                        loc=tuple((*loc, self.name)),
                    )
                ]
            )

    # These are implemented to work around pecularities of hashing bound methods on Python 3.7
    def __hash__(self) -> int:
        return id(self)  # pragma: no covet

    def __eq__(self, __o: object) -> bool:
        return self is __o  # pragma: no covet


@dataclass(frozen=True)
class FormFieldBodyExtractorMarker(BodyExtractorMarker):
    alias: typing.Optional[str]
    style: str
    explode: bool

    def register_parameter(self, param: inspect.Parameter) -> BodyExtractor:
        name = self.alias or param.name
        extractor = get_extractor(
            style=self.style, explode=self.explode, field=model_field_from_param(param)
        )
        return FormFieldBodyExtractor(
            alias=self.alias,
            style=self.style,
            explode=self.explode,
            name=name,
            extractor=extractor,
        )
